@@ -61,15 +61,17 @@
 (declare-function agent-shell-ui-backward-block "agent-shell")
 (declare-function agent-shell-ui-forward-block "agent-shell")
 (declare-function agent-shell-ui-mode "agent-shell")
+(declare-function agent-shell-completion-mode "agent-shell-completion")
 
 (defvar agent-shell-header-style)
 (defvar agent-shell-prefer-viewport-interaction)
 (defvar agent-shell-preferred-agent-config)
 
-(cl-defun agent-shell-viewport--show-buffer (&key text submit no-focus shell-buffer)
+(cl-defun agent-shell-viewport--show-buffer (&key append override submit no-focus shell-buffer)
   "Show a viewport compose buffer for the agent shell.
 
-TEXT is inserted into the viewport compose buffer.
+APPEND is appended to the viewport compose buffer.
+OVERRIDE, when non-nil, replaces content verbatim (no trimming).
 SUBMIT, when non-nil, submits after insertion.
 NO-FOCUS, when non-nil, avoids focusing the viewport compose buffer.
 SHELL-BUFFER, when non-nil, prefer this shell buffer.
@@ -84,44 +86,73 @@ Returns an alist with insertion details or nil otherwise:
     (error "Not yet supported"))
   (when no-focus
     (error "Not yet supported"))
+  (when (and append override)
+    (error "Use :append or :override but not both"))
   (when shell-buffer
     ;; Momentarily set buffer to same window, so it's recent in stack.
     (let ((current (current-buffer)))
       (pop-to-buffer-same-window shell-buffer)
       (pop-to-buffer-same-window current)))
-  (when-let ((shell-buffer (or shell-buffer (agent-shell--shell-buffer)))
-             (viewport-buffer (agent-shell-viewport--buffer :shell-buffer shell-buffer))
-             (text (or text (agent-shell--context) "")))
+  (when-let* ((shell-buffer (or shell-buffer (agent-shell--shell-buffer)))
+              (viewport-buffer (agent-shell-viewport--buffer :shell-buffer shell-buffer))
+              (text (or append (agent-shell--context :shell-buffer shell-buffer) "")))
+    (when (and override (not (string-empty-p text)))
+      (error "Cannot override"))
     (let ((insert-start nil)
           (insert-end nil))
       ;; Is there text to be inserted? Reject while busy.
       (when (and (agent-shell-viewport--busy-p
                   :viewport-buffer viewport-buffer)
-                 (not (string-empty-p (string-trim text))))
+                 (or (not (string-empty-p (string-trim text)))
+                     (and override (not (string-empty-p (string-trim override))))))
         (user-error "Busy... please wait"))
       (agent-shell--display-buffer viewport-buffer)
+      (when (and override
+                 (with-current-buffer viewport-buffer
+                   ;; viewport buffer empty?
+                   (not (= (buffer-size) 0))))
+        (unless (y-or-n-p "Compose buffer is not empty.  Override?")
+          ;; User does not want to override.
+          ;; Treat as regurlar text (typically appended).
+          (setq text (concat text
+                             (unless (string-empty-p text)
+                               "\n\n")
+                             override))
+          (setq override nil)))
       ;; TODO: Do we need to get prompt and partial response,
       ;; in case viewport compose buffer is created for the
       ;; first time on an ongoing/busy shell session?
-      (if (agent-shell-viewport--busy-p)
-          (agent-shell-viewport-view-mode)
-        (if (derived-mode-p 'agent-shell-viewport-edit-mode)
-            (unless (string-empty-p text)
-              (save-excursion
-                (goto-char (point-max))
-                (setq insert-start (point))
-                (unless (string-empty-p text)
-                  (insert "\n\n" text))
-                (setq insert-end (point))))
-          (agent-shell-viewport-edit-mode)
-          ;; Transitioned to edit mode. Start empty.
-          (agent-shell-viewport--initialize)
+      (cond
+       ((agent-shell-viewport--busy-p)
+        (agent-shell-viewport-view-mode))
+       (override
+        (agent-shell-viewport-edit-mode)
+        (agent-shell-viewport--initialize)
+        (setq insert-start (point))
+        (insert override)
+        (setq insert-end (point)))
+       ((derived-mode-p 'agent-shell-viewport-edit-mode)
+        (unless (string-empty-p text)
           (save-excursion
             (goto-char (point-max))
             (setq insert-start (point))
-            (unless (string-empty-p text)
-              (insert "\n\n" text))
+            (insert "\n\n" text)
             (setq insert-end (point)))))
+       (t
+        (agent-shell-viewport-edit-mode)
+        ;; Transitioned to edit mode. Wipe content.
+        (agent-shell-viewport--initialize)
+        ;; Restore snapshot if needed.
+        (when-let ((snapshot agent-shell-viewport--compose-snapshot))
+          (insert (map-elt snapshot :content))
+          (goto-char (map-elt snapshot :location))
+          (setq agent-shell-viewport--compose-snapshot nil))
+        (save-excursion
+          (goto-char (point-max))
+          (setq insert-start (point))
+          (unless (string-empty-p text)
+            (insert "\n\n" text))
+          (setq insert-end (point)))))
       `((:buffer . ,viewport-buffer)
         (:start . ,insert-start)
         (:end . ,insert-end)))))
@@ -131,6 +162,7 @@ Returns an alist with insertion details or nil otherwise:
   (interactive)
   (unless (derived-mode-p 'agent-shell-viewport-edit-mode)
     (user-error "Not in a shell viewport buffer"))
+  (setq agent-shell-viewport--compose-snapshot nil)
   (if agent-shell-prefer-viewport-interaction
       (agent-shell-viewport-compose-send-and-wait-for-response)
     (agent-shell-viewport-compose-send-and-kill)))
@@ -244,6 +276,7 @@ Optionally set its PROMPT and RESPONSE."
       (markdown-overlays-put))))
 
 (defun agent-shell-viewport--ensure-buffer ()
+  "Ensure current buffer is a viewport and err otherwise."
   (unless (or (derived-mode-p 'agent-shell-viewport-view-mode)
               (derived-mode-p 'agent-shell-viewport-edit-mode))
     (user-error "Not in a shell viewport buffer")))
@@ -300,14 +333,34 @@ Optionally set its PROMPT and RESPONSE."
   "Cancel prompt composition."
   (interactive)
   (agent-shell-viewport--ensure-buffer)
-  (if (or (derived-mode-p 'agent-shell-viewport-view-mode)
-          (with-current-buffer (agent-shell-viewport--shell-buffer)
-            (not (shell-maker-history))))
-      (bury-buffer)
-    ;; Edit mode
-    (when (or (string-empty-p (string-trim (buffer-string)))
-              (y-or-n-p "Discard composed prompt? "))
-      (agent-shell-viewport-view-last))))
+  (setq agent-shell-viewport--compose-snapshot nil)
+  (let ((viewport-buffer (current-buffer))
+        (shell-buffer (agent-shell-viewport--shell-buffer)))
+    ;; View mode
+    (if (or (derived-mode-p 'agent-shell-viewport-view-mode)
+            (with-current-buffer shell-buffer
+              (not (shell-maker-history))))
+        (bury-buffer)
+      ;; Edit mode
+      (when (or (string-empty-p (string-trim (buffer-string)))
+                (y-or-n-p "Discard composed prompt? "))
+        (if agent-shell-prefer-viewport-interaction
+            (agent-shell-viewport-view-last)
+          (agent-shell-other-buffer)
+          (kill-buffer viewport-buffer))))))
+
+(defun agent-shell-viewport-compose-peek-last ()
+  "Save compose buffer snapshot and peek at the last interaction."
+  (interactive)
+  (unless (derived-mode-p 'agent-shell-viewport-edit-mode)
+    (user-error "Not in a prompt compose buffer"))
+  (unless (with-current-buffer (agent-shell-viewport--shell-buffer)
+            (shell-maker-history))
+    (user-error "No items in history"))
+  (setq agent-shell-viewport--compose-snapshot
+        `((:content . ,(buffer-string))
+          (:location . ,(point))))
+  (agent-shell-viewport-view-last))
 
 (defun agent-shell-viewport-view-last ()
   "Display the last request/response interaction."
@@ -420,7 +473,7 @@ With EXISTING-ONLY, only return existing buffers without creating."
   (when-let ((shell-buffer (or shell-buffer
                                (agent-shell--shell-buffer))))
     (with-current-buffer shell-buffer
-      (let* ((viewport-buffer-name (concat (buffer-name shell-buffer)
+      (let* ((viewport-buffer-name (concat (buffer-name (get-buffer shell-buffer))
                                            agent-shell-viewport--suffix))
              (viewport-buffer (get-buffer viewport-buffer-name)))
         (if viewport-buffer
@@ -448,13 +501,18 @@ With EXISTING-ONLY, only return existing buffers without creating."
                                "\n\n"))))
     (with-current-buffer (agent-shell-viewport--shell-buffer)
       (goto-char (point-max)))
-    (agent-shell-viewport-edit-mode)
-    (if block-quoted-text
-        (progn
-          (agent-shell-viewport--initialize :prompt block-quoted-text)
-          (goto-char (point-max)))
+    (let ((snapshot agent-shell-viewport--compose-snapshot))
+      (agent-shell-viewport-edit-mode)
       (agent-shell-viewport--initialize)
-      (goto-char (point-min)))
+      (when snapshot
+        (insert (map-elt snapshot :content))
+        (setq agent-shell-viewport--compose-snapshot nil))
+      (when block-quoted-text
+        (goto-char (point-max))
+        (insert (if snapshot "\n\n" "") block-quoted-text))
+      (goto-char (if (or snapshot block-quoted-text)
+                     (point-max)
+                   (point-min))))
     ;; Setting point isn't enough at times. Force scrolling.
     (set-window-start (selected-window) (point-min))))
 
@@ -467,34 +525,49 @@ With EXISTING-ONLY, only return existing buffers without creating."
   "Show next interaction (request / response).
 
 If BACKWARDS is non-nil, go to previous interaction.
-If START-AT-TOP is non-nil, position at point-min regardless of direction."
+If START-AT-TOP is non-nil, position at point-min regardless of direction.
+
+If there are no more next items and a compose snapshot exists, restore the
+buffer from the snapshot and switch to edit mode."
   (interactive)
   (unless (derived-mode-p 'agent-shell-viewport-view-mode)
     (error "Not in a viewport buffer"))
   (when (agent-shell-viewport--busy-p)
     (user-error "Busy... please wait"))
-  (when-let ((shell-buffer (agent-shell-viewport--shell-buffer))
-             (viewport-buffer (current-buffer))
-             (next (with-current-buffer shell-buffer
-                     (if backwards
-                         (when (save-excursion
-                                 (let ((orig-line (line-number-at-pos)))
-                                   (comint-previous-prompt 1)
-                                   (= orig-line (line-number-at-pos))))
-                           (error "No previous page"))
-                       (when (save-excursion
-                               (let ((orig-line (point)))
-                                 (comint-next-prompt 1)
-                                 (= orig-line (point))))
-                         (error "No next page")))
-                     (shell-maker-next-command-and-response backwards))))
-    (agent-shell-viewport--initialize
-     :prompt (car next) :response (cdr next))
-    (goto-char (if start-at-top
-                   (point-min)
-                 (if backwards (point-max) (point-min))))
-    (agent-shell-viewport--update-header)
-    next))
+  (let ((shell-buffer (agent-shell-viewport--shell-buffer))
+        (viewport-buffer (current-buffer))
+        (snapshot agent-shell-viewport--compose-snapshot)
+        (pos (agent-shell-viewport--position :force-refresh t)))
+    ;; Check if at last position going forward with a snapshot to restore
+    (if (and (not backwards) snapshot pos
+             (= (car pos) (cdr pos)))
+        (progn
+          (agent-shell-viewport-edit-mode)
+          (agent-shell-viewport--initialize)
+          (insert (map-elt snapshot :content))
+          (goto-char (map-elt snapshot :location))
+          (setq agent-shell-viewport--compose-snapshot nil)
+          (cl-return-from agent-shell-viewport-next-page))
+      (when-let ((next (with-current-buffer shell-buffer
+                         (if backwards
+                             (when (save-excursion
+                                     (let ((orig-line (line-number-at-pos)))
+                                       (comint-previous-prompt 1)
+                                       (= orig-line (line-number-at-pos))))
+                               (error "No previous page"))
+                           (when (save-excursion
+                                   (let ((orig-line (point)))
+                                     (comint-next-prompt 1)
+                                     (= orig-line (point))))
+                             (error "No next page")))
+                         (shell-maker-next-command-and-response backwards))))
+        (agent-shell-viewport--initialize
+         :prompt (car next) :response (cdr next))
+        (goto-char (if start-at-top
+                       (point-min)
+                     (if backwards (point-max) (point-min))))
+        (agent-shell-viewport--update-header)
+        next))))
 
 (defun agent-shell-viewport-set-session-model ()
   "Set session model."
@@ -599,7 +672,11 @@ Automatically determines qualifier and bindings based on current major mode."
                       `((:key . ,(key-description (where-is-internal
                                                    'agent-shell-viewport-compose-cancel
                                                    agent-shell-viewport-edit-mode-map t)))
-                        (:description . "cancel"))))
+                        (:description . "cancel"))
+                      `((:key . ,(key-description (where-is-internal
+                                                   'agent-shell-viewport-compose-peek-last
+                                                   agent-shell-viewport-edit-mode-map t)))
+                        (:description . "previous page"))))
                     ((derived-mode-p 'agent-shell-viewport-view-mode)
                      (append
                       (list
@@ -657,6 +734,12 @@ Returns nil if VIEWPORT-BUFFER is not a viewport buffer or shell doesn't exist."
 (defvar-local agent-shell-viewport--position-cache nil
   "Cached position value (CURRENT . TOTAL).")
 
+(defvar-local agent-shell-viewport--compose-snapshot nil
+  "Alist with :content and :location from compose buffer before viewing history.")
+;; The viewport buffer transitions between major modes which clears
+;; buffer-local vars. Make snapshot permanent-local.
+(put 'agent-shell-viewport--compose-snapshot 'permanent-local t)
+
 (defun agent-shell-viewport--clean-up ()
   "Clean up resources.
 
@@ -684,6 +767,7 @@ For example, offer to kill associated shell session."
 (defvar agent-shell-viewport-edit-mode-map
   (let ((map (make-sparse-keymap)))
     (define-key map (kbd "C-c C-c") #'agent-shell-viewport-compose-send)
+    (define-key map (kbd "C-c C-p") #'agent-shell-viewport-compose-peek-last)
     (define-key map (kbd "C-c C-k") #'agent-shell-viewport-compose-cancel)
     (define-key map (kbd "C-<tab>") #'agent-shell-viewport-cycle-session-mode)
     (define-key map (kbd "C-c C-m") #'agent-shell-viewport-set-session-mode)
@@ -698,6 +782,8 @@ For example, offer to kill associated shell session."
 \\{agent-shell-viewport-edit-mode-map}"
   (cursor-intangible-mode +1)
   (setq buffer-read-only nil)
+  (when agent-shell-file-completion-enabled
+    (agent-shell-completion-mode +1))
   (agent-shell-viewport--update-header)
   (let ((inhibit-read-only t))
     (erase-buffer))
@@ -706,10 +792,10 @@ For example, offer to kill associated shell session."
 (defvar agent-shell-viewport-view-mode-map
   (let ((map (make-sparse-keymap)))
     (define-key map (kbd "C-c C-c") #'agent-shell-viewport-interrupt)
-    (define-key map (kbd "n") #'agent-shell-viewport-next-item)
-    (define-key map (kbd "p") #'agent-shell-viewport-previous-item)
     (define-key map (kbd "TAB") #'agent-shell-viewport-next-item)
     (define-key map (kbd "<backtab>") #'agent-shell-viewport-previous-item)
+    (define-key map (kbd "n") #'agent-shell-viewport-next-item)
+    (define-key map (kbd "p") #'agent-shell-viewport-previous-item)
     (define-key map (kbd "f") #'agent-shell-viewport-next-page)
     (define-key map (kbd "b") #'agent-shell-viewport-previous-page)
     (define-key map (kbd "r") #'agent-shell-viewport-reply)
